@@ -19,7 +19,7 @@
  *                                                                        *
  **************************************************************************/
 
-#include "proto.h"
+#include "prototypes.h"
 
 #include <string.h>
 
@@ -556,13 +556,21 @@ void do_scroll_down(void)
 
 	if (editwinrows > 1 && (openfile->edittop->next != NULL
 #ifndef NANO_TINY
-				|| chunk_for(openfile->firstcolumn, openfile->edittop) <
-					number_of_chunks_in(openfile->edittop)
+				|| (ISSET(SOFTWRAP) && (extra_chunks_in(openfile->edittop) >
+					chunk_for(openfile->firstcolumn, openfile->edittop)))
 #endif
 										))
 		edit_scroll(FORWARD);
 }
-#endif
+
+/* Scroll the line with the cursor to the center of the screen. */
+void do_center(void)
+{
+	adjust_viewport(CENTERING);
+	draw_all_subwindows();
+	full_refresh();
+}
+#endif /* !NANO_TINY || ENABLE_HELP */
 
 /* Move left one character. */
 void do_left(void)
