@@ -19,7 +19,7 @@
  *                                                                        *
  **************************************************************************/
 
-#include "proto.h"
+#include "prototypes.h"
 
 #include <errno.h>
 #ifdef HAVE_PWD_H
@@ -38,7 +38,7 @@ void get_homedir(void)
 #ifdef HAVE_PWD_H
 		/* When HOME isn't set, or when we're root, get the home directory
 		 * from the password file instead. */
-		if (homenv == NULL || geteuid() == 0) {
+		if (homenv == NULL || geteuid() == ROOT_UID) {
 			const struct passwd *userage = getpwuid(geteuid());
 
 			if (userage != NULL)
@@ -160,7 +160,7 @@ bool parse_line_column(const char *str, ssize_t *line, ssize_t *column)
 }
 
 /* In the given string, recode each embedded NUL as a newline. */
-void unsunder(char *string, size_t length)
+void recode_NUL_to_LF(char *string, size_t length)
 {
 	while (length > 0) {
 		if (*string == '\0')
@@ -171,7 +171,7 @@ void unsunder(char *string, size_t length)
 }
 
 /* In the given string, recode each embedded newline as a NUL. */
-void sunder(char *string)
+void recode_LF_to_NUL(char *string)
 {
 	while (*string != '\0') {
 		if (*string == '\n')
@@ -209,8 +209,8 @@ bool is_separate_word(size_t position, size_t length, const char *buf)
 	/* If the word starts at the beginning of the line OR the character before
 	 * the word isn't a letter, and if the word ends at the end of the line OR
 	 * the character after the word isn't a letter, we have a whole word. */
-	return ((position == 0 || !is_alpha_mbchar(before)) &&
-				(buf[word_end] == '\0' || !is_alpha_mbchar(after)));
+	return ((position == 0 || !is_alpha_char(before)) &&
+				(buf[word_end] == '\0' || !is_alpha_char(after)));
 }
 #endif /* ENABLE_SPELLER */
 
@@ -324,12 +324,14 @@ char *mallocstrcpy(char *dest, const char *src)
 	return dest;
 }
 
-/* Return an allocated copy of the first count characters of the given string. */
+/* Return an allocated copy of the first count characters
+ * of the given string, and NUL-terminate the copy. */
 char *measured_copy(const char *string, size_t count)
 {
-	char *thecopy = charalloc(count);
+	char *thecopy = charalloc(count + 1);
 
 	strncpy(thecopy, string, count);
+	thecopy[count] = '\0';
 
 	return thecopy;
 }
@@ -337,7 +339,7 @@ char *measured_copy(const char *string, size_t count)
 /* Return an allocated copy of the given string. */
 char *copy_of(const char *string)
 {
-	return measured_copy(string, strlen(string) + 1);
+	return measured_copy(string, strlen(string));
 }
 
 /* Free the string at dest and return the string at src. */
@@ -448,35 +450,35 @@ void remove_magicline(void)
 #endif
 
 #ifndef NANO_TINY
-/* Return in (top, top_x) and (bot, bot_x) the start and end "coordinates"
- * of the marked region.  If right_side_up isn't NULL, set it to TRUE when
- * the mark is at the top of the marked region, and to FALSE otherwise. */
-void get_region(const linestruct **top, size_t *top_x,
-				const linestruct **bot, size_t *bot_x, bool *right_side_up)
+/* Return TRUE when the mark is before or at the cursor, and FALSE otherwise. */
+bool mark_is_before_cursor(void)
 {
-	if (openfile->mark->lineno < openfile->current->lineno ||
-				(openfile->mark == openfile->current &&
-				openfile->mark_x < openfile->current_x)) {
+	return (openfile->mark->lineno < openfile->current->lineno ||
+						(openfile->mark == openfile->current &&
+						openfile->mark_x <= openfile->current_x));
+}
+
+/* Return in (top, top_x) and (bot, bot_x) the start and end "coordinates"
+ * of the marked region. */
+void get_region(linestruct **top, size_t *top_x, linestruct **bot, size_t *bot_x)
+{
+	if (mark_is_before_cursor()) {
 		*top = openfile->mark;
 		*top_x = openfile->mark_x;
 		*bot = openfile->current;
 		*bot_x = openfile->current_x;
-		if (right_side_up != NULL)
-			*right_side_up = TRUE;
 	} else {
 		*bot = openfile->mark;
 		*bot_x = openfile->mark_x;
 		*top = openfile->current;
 		*top_x = openfile->current_x;
-		if (right_side_up != NULL)
-			*right_side_up = FALSE;
 	}
 }
 
 /* Get the set of lines to work on -- either just the current line, or the
  * first to last lines of the marked region.  When the cursor (or mark) is
  * at the start of the last line of the region, exclude that line. */
-void get_range(const linestruct **top, const linestruct **bot)
+void get_range(linestruct **top, linestruct **bot)
 {
 	if (!openfile->mark) {
 		*top = openfile->current;
@@ -484,7 +486,7 @@ void get_range(const linestruct **top, const linestruct **bot)
 	} else {
 		size_t top_x, bot_x;
 
-		get_region(top, &top_x, bot, &bot_x, NULL);
+		get_region(top, &top_x, bot, &bot_x);
 
 		if (bot_x == 0 && *bot != *top && !also_the_last)
 			*bot = (*bot)->prev;
@@ -510,19 +512,15 @@ linestruct *line_from_number(ssize_t number)
 #endif /* !NANO_TINY */
 
 /* Count the number of characters from begin to end, and return it. */
-size_t get_totsize(const linestruct *begin, const linestruct *end)
+size_t number_of_characters_in(const linestruct *begin, const linestruct *end)
 {
 	const linestruct *line;
-	size_t totsize = 0;
+	size_t count = 0;
 
 	/* Sum the number of characters (plus a newline) in each line. */
 	for (line = begin; line != end->next; line = line->next)
-		totsize += mbstrlen(line->data) + 1;
+		count += mbstrlen(line->data) + 1;
 
-	/* The last line of a file doesn't have a newline -- otherwise it
-	 * wouldn't be the last line -- so subtract 1 when at EOF. */
-	if (line == NULL)
-		totsize--;
-
-	return totsize;
+	/* Do not count the final newline. */
+	return (count - 1);
 }
